#include "clm_gpu.h"

#include <dlfcn.h>
#include <stdio.h>

typedef int (*clm_gpuInitFunc)(unsigned int);
typedef void (*clm_gpuDestroyFunc)();
typedef void (*clm_linearInitFunc)(clm_Linear *linear);
typedef void (*clm_linearForwardFunc)(clm_Linear *linear, unsigned int batchSize, clm_Matrix *inputs);
typedef void (*clm_linearBackpropFunc)(clm_Linear *linear, float learnRate, unsigned int batchSize, clm_Matrix *inputs, bool updateErrors, clm_Matrix *outputErrors);

static void *lib;
static clm_gpuInitFunc initFunc;
static clm_gpuDestroyFunc destroyFunc;
static clm_linearInitFunc linearInitFunc;
static clm_linearForwardFunc linearForwardFunc;
static clm_linearBackpropFunc linearBackpropFunc;

int clm_gpuInit(unsigned int mode) {
	const char *libName = NULL;
	if(mode & CLM_MODE_CPU) {
		libName = "libclm_cpu.so";
	} else if(mode & CLM_MODE_OPENCL) {
		libName = "libclm_opencl.so";
	} else {
		printf("No valid mode supplied. Must set either CLM_MODE_CPU or CLM_MODE_GPU");
		return 1;
	}

	lib = dlopen(libName, RTLD_NOW);
	if(!lib) {
		printf("Failed to load lib: %s\n", dlerror());
		return 1;
	}

	initFunc = (clm_gpuInitFunc) dlsym(lib, "clm_gpuInit");
	destroyFunc = (clm_gpuDestroyFunc) dlsym(lib, "clm_gpuDestroy");
	linearInitFunc = (clm_linearInitFunc) dlsym(lib, "clm_linearInit");
	linearForwardFunc = (clm_linearForwardFunc) dlsym(lib, "clm_linearForward");
	linearBackpropFunc = (clm_linearBackpropFunc) dlsym(lib, "clm_linearBackprop");

	if(!initFunc || !destroyFunc || !linearForwardFunc || !linearBackpropFunc) {
		printf("Failed to load functions from library: %s\n", dlerror());
		return 1;
	}

	return initFunc(mode);
}

void clm_gpuDestroy() {
	destroyFunc();
	dlclose(lib);
}

void clm_linearInit(clm_Linear *linear) {
	linearInitFunc(linear);
}

void clm_linearForward(clm_Linear *linear, unsigned int batchSize, clm_Matrix *inputs) {
	linearForwardFunc(linear, batchSize, inputs);
}

void clm_linearBackprop(clm_Linear *linear, float learnRate, unsigned int batchSize, clm_Matrix *inputs, bool updateErrors, clm_Matrix *outputErrors) {
	linearBackpropFunc(linear, learnRate, batchSize, inputs, updateErrors, outputErrors);
}
