#include "clm.h"
#include "clm_gpu.h"
#include <math.h>

#include "clm_util.h"

#include <CL/cl_platform.h>

#include <CL/cl.h>

#include <stdio.h>
#include <string.h>

static cl_context context;
static cl_device_id deviceID;
static cl_command_queue queue;
static cl_kernel kernelLinearForward;
static cl_kernel kernelLinearBackprop1;
static cl_kernel kernelLinearBackprop2;
static size_t kernelLinearForwardLocal;
static size_t kernelLinearBackprop1Local;
static size_t kernelLinearBackprop2Local;

struct clm_NativeBuf {
	cl_mem mem;
};

typedef struct __attribute__((packed)) {
	cl_uint rows;
	cl_uint cols;
	cl_char transposed;
} cl_GPUMat;

#define gpuMat(mat) \
	((cl_GPUMat){.rows = mat.rows, .cols = mat.cols, .transposed = mat.transposed})

int clm_gpuInit(unsigned int mode) {
	// Connect to a compute device
	int useGPU = true;
	cl_int err = clGetDeviceIDs(NULL, useGPU ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &deviceID, NULL);
	if(err != CL_SUCCESS) {
		printf("Error: Failed to create a device group\n");
		return 1;
	}

	char *buffer = clm_loadFile("src/mat.cl");
	if(!buffer) {
		printf("Failed to load mat kernel\n");
		return 1;
	}

	// printf("%s", buffer);

	context = clCreateContext(NULL, 1, &deviceID, NULL, NULL, &err);
	if(!context) {
		printf("Failed to create context\n");
		return 1;
	}

	queue = clCreateCommandQueueWithProperties(context, deviceID, NULL, &err);
	if(!queue) {
		printf("Failed to create command queue\n");
		return 1;
	}

	size_t length = strlen(buffer);
	cl_program program = clCreateProgramWithSource(context, 1, (const char **) &buffer, &length, &err);
	if(!program) {
		printf("Failed to create program\n");
		return 1;
	}

	err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to build program\n");
		// clGetProgramBuildInfo...
		return 1;
	}

	kernelLinearForward = clCreateKernel(program, "linear_forward", &err);
	kernelLinearBackprop1 = clCreateKernel(program, "linear_backprop_1", &err);
	kernelLinearBackprop2 = clCreateKernel(program, "linear_backprop_2", &err);
	if(!kernelLinearForward || !kernelLinearBackprop1 || !kernelLinearBackprop2) {
		printf("Failed to create kernels: %d\n", err);
		return 1;
	}

	err = clGetKernelWorkGroupInfo(kernelLinearForward, deviceID, CL_KERNEL_WORK_GROUP_SIZE, sizeof(kernelLinearForwardLocal), &kernelLinearForwardLocal, NULL);
	err |= clGetKernelWorkGroupInfo(kernelLinearBackprop1, deviceID, CL_KERNEL_WORK_GROUP_SIZE, sizeof(kernelLinearBackprop1Local), &kernelLinearBackprop1Local, NULL);
	err |= clGetKernelWorkGroupInfo(kernelLinearBackprop2, deviceID, CL_KERNEL_WORK_GROUP_SIZE, sizeof(kernelLinearBackprop2Local), &kernelLinearBackprop2Local, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to get work group sizes\n");
		return 1;
	}

	return 0;
}

void clm_gpuDestroy() {
}

static cl_mem allocGPUMat(cl_GPUMat mat, unsigned int count, cl_mem_flags flags, clm_NativeBuf *nativeBuf) {
	cl_int err;

	if(!nativeBuf->mem) {
		cl_mem mat_values = clCreateBuffer(context, flags, count * sizeof(float) * mat.rows * mat.cols, NULL, &err);
		if(!mat_values) {
			printf("Failed to alloc buffer: %s\n", clm_clErrorToString(err));
			return NULL;
		}
		nativeBuf->mem = mat_values;
	}

	return nativeBuf->mem;
}

static cl_mem writeGPUMat(cl_GPUMat gpuMat, clm_Matrix mat, cl_mem_flags flags, clm_NativeBuf *nativeBuf) {
	cl_int err;
	cl_mem mem = allocGPUMat(gpuMat, 1, flags, nativeBuf);

	err = clEnqueueWriteBuffer(queue, mem, CL_TRUE, 0, sizeof(float) * mat.rows * mat.cols, mat.values, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue write: %s\n", clm_clErrorToString(err));
		return NULL;
	}

	return mem;
}

static cl_mem writeGPUMats(cl_GPUMat gpuMat, unsigned int numMats, clm_Matrix *mats, cl_mem_flags flags, clm_NativeBuf *nativeBuf) {
	cl_int err;
	cl_mem mem = allocGPUMat(gpuMat, numMats, flags, nativeBuf);

	for(unsigned int i = 0; i < numMats; i++) {
		err = clEnqueueWriteBuffer(queue, mem, CL_FALSE, i * sizeof(float) * gpuMat.rows * gpuMat.cols, sizeof(float) * gpuMat.rows * gpuMat.cols, mats[i].values, 0, NULL, NULL);
		if(err != CL_SUCCESS) {
			printf("Failed to enqueue write: %s\n", clm_clErrorToString(err));
			return NULL;
		}
	}

	clFlush(queue);
	clFinish(queue);

	return mem;
}

static cl_mem writeNativeMatrix(clm_NativeMatrix matrix) {
	cl_mem mem = matrix.native->mem;
	cl_int err = clEnqueueWriteBuffer(queue, mem, CL_TRUE, 0, sizeof(float) * matrix.matrix.rows * matrix.matrix.cols, matrix.matrix.values, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue write: %s\n", clm_clErrorToString(err));
		return NULL;
	}

	return mem;
}

static cl_mem writeNativeMatrixArray(clm_NativeMatrixArray array, unsigned int n) {
	clm_Matrix mat = array.matrixes[0];
	cl_mem mem = array.native->mem;

	size_t matLength = sizeof(float) * mat.rows * mat.cols;
	for(unsigned int i = 0; i < array.length; i++) {
		cl_int err = clEnqueueWriteBuffer(queue, mem, CL_FALSE, i * matLength, matLength, array.matrixes[i].values, 0, NULL, NULL);
		if(err != CL_SUCCESS) {
			printf("Failed to enqueue write: %s\n", clm_clErrorToString(err));
			return NULL;
		}
	}

	clFlush(queue);
	clFinish(queue);

	return mem;
}

static void readGPUMats(cl_GPUMat mat, unsigned int numMats, clm_Matrix *mats, clm_NativeBuf *nativeBuf) {
	cl_int err;
	cl_mem mem = nativeBuf->mem;

	for(unsigned int i = 0; i < numMats; i++) {
		err = clEnqueueReadBuffer(queue, mem, CL_FALSE, i * sizeof(float) * mat.rows * mat.cols, sizeof(float) * mat.rows * mat.cols, mats[i].values, 0, NULL, NULL);
		if(err != CL_SUCCESS) {
			printf("Failed to enqueue read: %s\n", clm_clErrorToString(err));
			return;
		}
	}

	clFlush(queue);
	clFinish(queue);
}

static void readNativeMatrix(clm_NativeMatrix matrix) {
	cl_int err = clEnqueueReadBuffer(queue, matrix.native->mem, CL_TRUE, 0, sizeof(float) * matrix.matrix.rows * matrix.matrix.cols, matrix.matrix.values, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue read: %s\n", clm_clErrorToString(err));
		return;
	}
}

static void readNativeMatrixArray(clm_NativeMatrixArray array, unsigned int n) {
	clm_Matrix mat = array.matrixes[0];
	size_t matLength = sizeof(float) * mat.rows * mat.cols;

	// TODO: don't do blocking reads, instead wait once at the end
	for(unsigned int i = 0; i < n; i++) {
		cl_int err = clEnqueueReadBuffer(queue, array.native->mem, CL_FALSE, i * matLength, matLength, array.matrixes[i].values, 0, NULL, NULL);
		if(err != CL_SUCCESS) {
			printf("Failed to enqueue read: %s\n", clm_clErrorToString(err));
			return;
		}
	}

	clFlush(queue);
	clFinish(queue);
}

static void clm_nativeAllocMatrix(clm_NativeMatrix *matrix, cl_mem_flags flags) {
	matrix->native = calloc(1, sizeof(clm_NativeBuf));
	allocGPUMat(gpuMat(matrix->matrix), 1, flags, matrix->native);
}

static void clm_nativeAllocMatrixArray(clm_NativeMatrixArray *matrixArray, cl_mem_flags flags) {
	matrixArray->native = calloc(1, sizeof(clm_NativeBuf));
	allocGPUMat(gpuMat(matrixArray->matrixes[0]), matrixArray->length, flags, matrixArray->native);
}

void clm_linearInit(clm_Linear *linear) {
	clm_nativeAllocMatrix(&linear->weights, CL_MEM_READ_ONLY);
	clm_nativeAllocMatrix(&linear->bias, CL_MEM_READ_ONLY);
	clm_nativeAllocMatrixArray(&linear->output, CL_MEM_READ_WRITE);
	clm_nativeAllocMatrixArray(&linear->error, CL_MEM_READ_WRITE);
	clm_nativeAllocMatrixArray(&linear->weightsError, CL_MEM_READ_WRITE);
	clm_nativeAllocMatrixArray(&linear->gradient, CL_MEM_READ_WRITE);

	// TODO: alloc mem
	linear->nativeInput = calloc(1, sizeof(clm_NativeBuf));
	linear->nativeOutputErrors = calloc(1, sizeof(clm_NativeBuf));
}

// TODO: allow writing multiple inputs at once to improve throughput (no need to rewrite weights/bias each time)
void clm_linearForward(clm_Linear *linear, unsigned int batchSize, clm_Matrix *inputs) {
	if(batchSize == 0) return;

	cl_GPUMat matInput = gpuMat(inputs[0]);
	cl_GPUMat matWeights = gpuMat(linear->weights.matrix);
	cl_GPUMat matBias = gpuMat(linear->bias.matrix);
	cl_GPUMat matOut = gpuMat(linear->output.matrixes[0]);

	size_t workSize = matOut.rows * matOut.cols;

	cl_int err;

	// TODO: make sure to always alloc nn.batchSize, not batchSize
	cl_mem matInput_values = writeGPUMats(matInput, batchSize, inputs, CL_MEM_READ_ONLY, linear->nativeInput);
	cl_mem matWeights_values = writeNativeMatrix(linear->weights);
	cl_mem matBias_values = writeNativeMatrix(linear->bias);
	if(!matInput_values || !matWeights_values || !matBias_values) {
		// linear->output = INVALID_MATRIX;
		return;
	}

	cl_mem matOut_values = linear->output.native->mem;

	err = 0;
	err |= clSetKernelArg(kernelLinearForward, 0, sizeof(cl_uint), &batchSize);
	err |= clSetKernelArg(kernelLinearForward, 1, sizeof(matInput), &matInput);
	err |= clSetKernelArg(kernelLinearForward, 2, sizeof(matInput_values), &matInput_values);
	err |= clSetKernelArg(kernelLinearForward, 3, sizeof(matWeights), &matWeights);
	err |= clSetKernelArg(kernelLinearForward, 4, sizeof(matWeights_values), &matWeights_values);
	err |= clSetKernelArg(kernelLinearForward, 5, sizeof(matBias), &matBias);
	err |= clSetKernelArg(kernelLinearForward, 6, sizeof(matBias_values), &matBias_values);
	err |= clSetKernelArg(kernelLinearForward, 7, sizeof(matOut), &matOut);
	err |= clSetKernelArg(kernelLinearForward, 8, sizeof(matOut_values), &matOut_values);
	if(err != CL_SUCCESS) {
		printf("Failed to set kernel args (1): %d\n", err);
		return;
	}

	/*char *info = calloc(1, 1024);
	clGetProgramInfo(program, CL_PROGRAM_STRING_DEBUG_INFO, 1024, info, NULL);
	printf("INFO: %s\n", info);*/

	size_t global = ceil((float) workSize / kernelLinearForwardLocal) * kernelLinearForwardLocal;
	err = clEnqueueNDRangeKernel(queue, kernelLinearForward, 1, NULL, &global, &kernelLinearForwardLocal, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue: %d\n", err);
		return;
	}

	clFlush(queue);
	clFinish(queue);

	readNativeMatrixArray(linear->output, batchSize);
}

void clm_linearBackprop(clm_Linear *linear, float learnRate, unsigned int batchSize, clm_Matrix *inputs, bool updateErrors, clm_Matrix *outputErrors) {
	if(batchSize == 0) return;

	cl_GPUMat matInput = gpuMat(inputs[0]);
	cl_GPUMat matWeights = gpuMat(linear->weights.matrix);
	cl_GPUMat matOutput = gpuMat(linear->output.matrixes[0]);
	cl_GPUMat matInputErrors = gpuMat(linear->error.matrixes[0]);
	cl_GPUMat matOutputErrors = !updateErrors ? (cl_GPUMat){0} : (cl_GPUMat) gpuMat(outputErrors[0]);
	cl_GPUMat matOutputWeightsErrors = gpuMat(linear->weightsError.matrixes[0]);
	cl_GPUMat matOutputGradients = gpuMat(linear->gradient.matrixes[0]);

	cl_int err;

	cl_mem matInput_values = writeGPUMats(matInput, batchSize, inputs, CL_MEM_READ_ONLY, linear->nativeInput);
	cl_mem matWeights_values = writeNativeMatrix(linear->weights);
	cl_mem matInputErrors_values = writeNativeMatrixArray(linear->error, batchSize);
	cl_mem matOutput_values = writeNativeMatrixArray(linear->output, batchSize);
	if(!matInput_values || !matWeights_values || !matInputErrors_values || !matOutput_values) {
		printf("Failed to write GPU mats\n");
		return;
	}

	// cl_mem matOutput_values = allocGPUMat(matOutput, batchSize, CL_MEM_READ_ONLY, linear->nativeOutput);
	cl_mem matOutputErrors_values = !updateErrors ? NULL : allocGPUMat(matOutputErrors, batchSize, CL_MEM_READ_WRITE, linear->nativeOutputErrors);
	cl_mem matOutputWeightsErrors_values = linear->weightsError.native->mem;
	cl_mem matOutputGradients_values = linear->gradient.native->mem;
	if(!matOutputWeightsErrors_values || !matOutputGradients_values) {
		printf("Failed to alloc GPU mats\n");
		return;
	}

	err = 0;
	err |= clSetKernelArg(kernelLinearBackprop1, 0, sizeof(cl_uint), &batchSize);
	err |= clSetKernelArg(kernelLinearBackprop1, 1, sizeof(cl_float), &learnRate);
	err |= clSetKernelArg(kernelLinearBackprop1, 2, sizeof(matWeights), &matWeights);
	err |= clSetKernelArg(kernelLinearBackprop1, 3, sizeof(matWeights_values), &matWeights_values);
	err |= clSetKernelArg(kernelLinearBackprop1, 4, sizeof(matOutput), &matOutput);
	err |= clSetKernelArg(kernelLinearBackprop1, 5, sizeof(matOutput_values), &matOutput_values);
	err |= clSetKernelArg(kernelLinearBackprop1, 6, sizeof(matInputErrors), &matInputErrors);
	err |= clSetKernelArg(kernelLinearBackprop1, 7, sizeof(matInputErrors_values), &matInputErrors_values);
	err |= clSetKernelArg(kernelLinearBackprop1, 8, sizeof(matOutputGradients), &matOutputGradients);
	err |= clSetKernelArg(kernelLinearBackprop1, 9, sizeof(matOutputGradients_values), &matOutputGradients_values);
	if(err != CL_SUCCESS) {
		printf("Failed to set kernel args (2): %s\n", clm_clErrorToString(err));
		return;
	}

	size_t step1WorkSize = matOutputGradients.rows * matOutputGradients.cols;
	size_t step1Global = ceil((float) step1WorkSize / kernelLinearBackprop1Local) * kernelLinearBackprop1Local;
	err = clEnqueueNDRangeKernel(queue, kernelLinearBackprop1, 1, NULL, &step1Global, &kernelLinearBackprop1Local, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue: %s\n", clm_clErrorToString(err));
		return;
	}

	clFlush(queue);
	clFinish(queue);

	err = 0;
	err |= clSetKernelArg(kernelLinearBackprop2, 0, sizeof(cl_uint), &batchSize);
	err |= clSetKernelArg(kernelLinearBackprop2, 1, sizeof(matWeights), &matWeights);
	err |= clSetKernelArg(kernelLinearBackprop2, 2, sizeof(matWeights_values), &matWeights_values);
	err |= clSetKernelArg(kernelLinearBackprop2, 3, sizeof(matInput), &matInput);
	err |= clSetKernelArg(kernelLinearBackprop2, 4, sizeof(matInput_values), &matInput_values);
	err |= clSetKernelArg(kernelLinearBackprop2, 5, sizeof(matInputErrors), &matInputErrors);
	err |= clSetKernelArg(kernelLinearBackprop2, 6, sizeof(matInputErrors_values), &matInputErrors_values);
	err |= clSetKernelArg(kernelLinearBackprop2, 7, sizeof(cl_char), &updateErrors);

	err |= clSetKernelArg(kernelLinearBackprop2, 8, sizeof(matOutputErrors), &matOutputErrors);
	if(updateErrors) {
		err |= clSetKernelArg(kernelLinearBackprop2, 9, sizeof(matOutputErrors_values), &matOutputErrors_values);
	} else {
		err |= clSetKernelArg(kernelLinearBackprop2, 9, sizeof(matOutputErrors_values), NULL);
	}

	err |= clSetKernelArg(kernelLinearBackprop2, 10, sizeof(matOutputWeightsErrors), &matOutputWeightsErrors);
	err |= clSetKernelArg(kernelLinearBackprop2, 11, sizeof(matOutputWeightsErrors_values), &matOutputWeightsErrors_values);
	err |= clSetKernelArg(kernelLinearBackprop2, 12, sizeof(matOutputGradients), &matOutputGradients);
	err |= clSetKernelArg(kernelLinearBackprop2, 13, sizeof(matOutputGradients_values), &matOutputGradients_values);
	if(err != CL_SUCCESS) {
		printf("Failed to set kernel args (3): %s\n", clm_clErrorToString(err));
		return;
	}

	size_t step2WorkSize = matOutputWeightsErrors.rows * matOutputWeightsErrors.cols;
	size_t step2Global = ceil((float) step2WorkSize / kernelLinearBackprop2Local) * kernelLinearBackprop2Local;
	err = clEnqueueNDRangeKernel(queue, kernelLinearBackprop2, 1, NULL, &step2Global, &kernelLinearBackprop2Local, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue: %s\n", clm_clErrorToString(err));
		return;
	}

	clFlush(queue);
	clFinish(queue);

	readNativeMatrixArray(linear->weightsError, batchSize);
	readNativeMatrixArray(linear->gradient, batchSize);

	if(updateErrors) readGPUMats(matOutputErrors, batchSize, outputErrors, linear->nativeOutputErrors);
}
